#!/usr/bin/python

import eon_read
import optparse
import os
from optparse import OptionParser
import subprocess
import shutil
import write_statelist
import numpy
from collections import deque
import sys 
from time import sleep

parser = OptionParser()
parser.add_option("-d", "--dir",type="string", dest="in_dir", help="Path of the directory eon results", metavar="DIR",default=os.getcwd())
parser.add_option("--eondir",type="string", dest="eondir", help ="Directory containing eon with hessianjob")
parser.add_option("-n","--ncpu",type="int",dest="ncpu",help="Number of CPUs",default=1)
(options, args) = parser.parse_args()



if not options.in_dir:
    print "ERROR: specify input and output directories"
    exit(0)

in_path=    os.path.join(os.getcwd(),options.in_dir)
eon_path =  os.path.join(os.getcwd(),options.eondir)

if not os.path.isdir(in_path):
    print "ERROR: '%s' is not a valid directory" % in_path
    exit(0)
if not os.path.isdir(eon_path):
    print "ERROR: '%s' is not a valid directory" % eon_path
if not os.path.isdir(os.path.join(in_path,"states")):
    print "ERROR: '%s' is not a valid EON directory" % in_path
    exit(0)

#######################################################################################################


def find_reverse_proc(statelist,stateid,procid):
    reverse_proc_id=None
    eps = 0.000001
    proc = statelist[ stateid ]['proclist'][procid]
    product_id = proc['productid']

    #if the process is to itsself:
    if proc['reactantid'] == proc['productid']:
        for candidate in statelist[ product_id ]['proclist']:
            if 'calculated_prefactor' in candidate:
                continue
            if ( abs(1.0 - candidate['saddle_energy']/proc['saddle_energy'] ) < eps ):
                reverse_proc_id = candidate['procid']
                break
    else:
        for candidate in statelist[ product_id ]['proclist']:
            if candidate['productid'] != stateid:
                continue
            #if the prefactor was already set for this process, it cannot count as the reverse
            if 'calculated_prefactor' in candidate:
                continue


            if(  ( abs(1.0 - candidate['prefactor']/proc['product_prefactor'] ) < eps ) and
                 ( abs(1.0 - candidate['product_prefactor']/proc['prefactor'] ) < eps ) and
                 ( abs(1.0 - candidate['saddle_energy']/proc['saddle_energy'] ) < eps )  ):

                reverse_proc_id = candidate['procid']
                break
    
    return reverse_proc_id

#######################################################################################################

def newjob(joblist,dirname,cpuid,config,hessiantype,stateid,procid=None):
    job = { }
    print "Starting job, dirname = %s" % dirname
    print "Stateid=%d, type = %s, process id = %s" % (stateid,hessiantype,repr(procid))
    print " "
    #copy config and configuration into place:
    if hessiantype == "saddle":
        filename = os.path.join(dirname,"saddle_passed.con")
        shutil.copyfile( os.path.join(dirname,"..","config_saddle.ini"),   os.path.join(dirname,"config_passed.ini") )
    elif hessiantype =="reactant":
        filename = os.path.join(dirname,"reactant_passed.con")
        shutil.copyfile( os.path.join(dirname,"..","config_reactant.ini"), os.path.join(dirname,"config_passed.ini") )
    else:
        return []

    confile = open(filename,'w')
    print "writing configuration to %s" % filename
    config.write_con(confile)
    confile.close()
    
    job['path']=dirname
    job['type']=hessiantype
    job['cpuid'] = cpuid
    job['stateid']=stateid
    job['procid']=procid
    job['process'] = subprocess.Popen(os.path.join(dirname,"client"),cwd = dirname)
    
    if len(joblist) == cpuid:
        joblist.append(job)
    else:
        joblist[cpuid] = job
    
    return

#######################################################################################################

def savemode(modefilename,modelist):
    print "saving modelist of length %d to %s" % (len(modelist),modefilename)
    modefile = eon_read._open_file_writing(modefilename)
    for frequency in modelist:
        modefile.write("%.6f\n" % float(frequency))
    modefile.close()            
            
#######################################################################################################

def finishjob(statelist,joblist,directories):
    '''
    Find the first job which finishes, registers the results and returns its cpuid so a new job can start.
    '''
    while True:
        AtLeastOneRunning=False
        for workingdir in directories: 
            if workingdir['busy'] == True:
                AtLeastOneRunning=True
                cpuid = workingdir['cpu']
                #poll process, if it is not running anymore register and return.
                if joblist[cpuid]['process'].poll() != None:
                    job = joblist[cpuid]
                    print "%s job in state %d (process %s) finished, registering" % (job['type'],job['stateid'],repr(job['procid']) )
                    saveresults(job,statelist,directories)
                    joblist[cpuid]=None
                    return cpuid
         #if no more process are running, return -1           
        if AtLeastOneRunning==False:
            break
        print "sleeping"
        sleep(5)

    return -1            

#######################################################################################################            

def saveresults(job,statelist,directories):
    assert(directories[ job['cpuid'] ]['path'] == job['path'] )
    directories[ job['cpuid'] ]['busy'] = False

    #analyze result
    
    if job['process'].returncode == 0:
        print "process finished"
        modefilename = os.path.join(job['path'],"mode.dat")
        modefile = open(modefilename,'r')
        modelist = []
        for line in modefile:
            modelist.append(float(line))
        modefile.close()
        
        #save exact modes
        if job['type']=="reactant":
            modefilename = os.path.join(in_path,"states",str(job['stateid']),"mode_exact.dat")
            savemode(modefilename,modelist)
            
        elif job['type']=="saddle":
            #maybe the process has already been registered, as reverse of some other process, in that case we continue
            if 'calculated_prefactor' in statelist[ job['stateid'] ]['proclist'][ job['procid']]:
                print "process was already saved as reverse process"
                return

            #find reverse process            
            productid = statelist[ job['stateid'] ]['proclist'][ job['procid'] ]['productid']
            reverse_proc_id = find_reverse_proc(statelist,job['stateid'],job['procid'])
            if reverse_proc_id==None:
                print "Could not find reverse process for proc %d from state %d" % (job['procid'],job['stateid'])
                exit(0)

            if 'calculated_prefactor' not in statelist[ job['stateid'] ]['proclist'][ job['procid']]:
                modefilename = os.path.join(in_path,"states",str(job['stateid']),"procdata","mode_exact_"+str(job['procid'])+".dat")
                savemode(modefilename,modelist)
                statelist[ job['stateid'] ]['proclist'][ job['procid'] ]['calculated_prefactor'] = True
            
            if 'calculated_prefactor' not in  statelist[productid]['proclist'][reverse_proc_id]:
                statelist[productid]['proclist'][reverse_proc_id]['calculated_prefactor'] = True
                modefilename = os.path.join(in_path,"states",str( productid ),"procdata","mode_exact_"+str(reverse_proc_id)+".dat")
                savemode(modefilename,modelist)
    else:
        print "ERROR: returncode was not zero"                

#######################################################################################################                      
    
def makedirs(eon_path,ncpus):
    cpuid = 0
    dirnames = []
    while cpuid < ncpus:
        dirdict = { }
        sys.stderr.write("making dir for cpu %d\n" % cpuid)
        jobdirname = os.path.join(eon_path,str(cpuid)) 
        os.mkdir(jobdirname)
        #copy necessary files there. 
        shutil.copyfile( os.path.join(eon_path,"reactant_passed.con"), os.path.join(jobdirname,"reactant_passed.con") )               
        shutil.copyfile( os.path.join(eon_path,"saddle_passed.con"), os.path.join(jobdirname,"saddle_passed.con") )               
        shutil.copyfile( os.path.join(eon_path,"product_passed.con"), os.path.join(jobdirname,"product_passed.con") )                       
        shutil.copy2( os.path.join(eon_path,"client"), os.path.join(jobdirname,"client") )                       
        dirdict['path'] = jobdirname
        dirdict['cpu'] = cpuid
        dirdict['busy'] = False
        dirnames.append(dirdict)
        cpuid+=1
    return dirnames        

#######################################################################################################
            
def main():
    statelist = eon_read.read_statelist(in_path,read_configurations=True,read_processes=True,read_reactant=True)
    

    joblist =[]
    directories = makedirs(eon_path,options.ncpu)


    istate = 0
    irunning = 0
    while istate < len(statelist):
        #if not all slots are filled:
        if irunning < options.ncpu:
            skip = False
            for job in joblist:
                if job!=None:
                    #if already calculating modes for this state, continue
                    if job['stateid']==istate and job['type']=="reactant":
                        skip = True            
            if skip == False:
                directories[irunning]['busy'] = True    
                newjob(joblist,directories[irunning]['path'],irunning,statelist[istate]['reactant_config'],"reactant",istate)
                irunning+=1

        #if all slots are filled, wait for first job to finish:
        else:
            cpuid = finishjob(statelist,joblist,directories)
            skip = False
            for job in joblist:
                if job!=None:
                    if job['stateid']==istate and job['type']=="reactant":
                        skip = True            
            if skip == False:
                directories[cpuid]['busy'] = True    
                newjob(joblist,directories[cpuid]['path'],cpuid,statelist[istate]['reactant_config'],"reactant",istate)

        numprocs = len(statelist[istate]['proclist'])
        iproc = 0
        while iproc < numprocs:
            if statelist[istate]['proclist'][iproc]['productid']==-1:
                iproc +=1
                continue
            if 'calculated_prefactor' in statelist[istate]['proclist'][iproc]:
                iproc +=1
                continue

            if irunning < options.ncpu:
                skip = False
                for job in joblist:
                    if job!=None:
                        if job['stateid']==istate and job['procid'] == iproc and job['type']=="saddle":
                            skip = True            
                if skip == False:
                    directories[irunning]['busy'] = True    
                    newjob(joblist,directories[irunning]['path'],irunning,statelist[istate]['proclist'][ iproc ]['saddle_config'],"saddle",istate,iproc)
                    irunning+=1                
            else:    
                
                cpuid = finishjob(statelist,joblist,directories)    
                skip = False
                for job in joblist:
                    if job!=None:
                        if job['stateid']==istate and job['procid'] == iproc and job['type']=="saddle":
                            skip = True                 
            
                if skip == False:
                    directories[cpuid]['busy'] = True    
                    newjob(joblist,directories[cpuid]['path'],cpuid,statelist[istate]['proclist'][ iproc ]['saddle_config'],"saddle",istate,iproc)            
            iproc+=1

        istate +=1            
    #finish all remaining jobs.
    while True:
        cpuid = finishjob(statelist,joblist,directories)
        if cpuid==-1:
            break
        
            
#######################################################################################################

if __name__ == "__main__":
    main()

